# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

import secrets
from unittest import mock

import pytest
import urllib3

import croud.config
from croud.api import Client
from croud.config.configuration import Configuration
from tests.util.fake_cloud import FakeCrateDBCloud

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


@pytest.fixture(scope="session")
def fake_cratedb_cloud():
    with FakeCrateDBCloud() as cloud:
        yield cloud


@pytest.fixture
def config(fake_cratedb_cloud, tmp_path):
    profile = secrets.token_urlsafe(16)

    config = Configuration("croud.yaml", tmp_path)
    config.add_profile(profile, endpoint=f"https://127.0.0.1:{fake_cratedb_cloud.port}")
    config.set_auth_token(profile, secrets.token_urlsafe(64))
    config.use_profile(profile)

    with mock.patch.object(croud.config, "_CONFIG", config):
        yield config


@pytest.fixture
def client(fake_cratedb_cloud, config):
    yield Client(
        config.endpoint,
        token=config.token,
        on_token=config.set_current_auth_token,
        region="bregenz.a1",
        _verify_ssl=False,
    )



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.


import pytest
import requests

from croud.server import Server


@pytest.mark.parametrize(
    "qs,status_code,message,token_value",
    [
        ("?token=foo", 200, "You have successfully logged into CrateDB Cloud!", "foo"),
        ("", 400, "Authentication token missing in URL.", None),
        (
            "?token=foo&token=bar",
            400,
            "ore than one authentication token present in URL.",
            None,
        ),
    ],
)
def test_token_handler_and_login(qs, status_code, message, token_value):
    token_store = {}

    def on_token(token):
        token_store["token"] = token

    server = Server(on_token, random_port=True).start_in_background()

    response = requests.get(f"http://localhost:{server.port}/{qs}")

    if token_value is not None:
        assert token_store["token"] == token_value
    else:
        assert "token" not in token_store
    assert response.status_code == status_code
    assert message in response.text



# PYTHON_ARGCOMPLETE_OK
#
# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

import sys
from distutils.util import strtobool

import colorama

from croud.clusters.commands import (
    clusters_delete,
    clusters_deploy,
    clusters_get,
    clusters_list,
    clusters_restart_node,
    clusters_scale,
    clusters_set_deletion_protection,
    clusters_upgrade,
)
from croud.config import CONFIG
from croud.config.commands import (
    config_add_profile,
    config_current_profile,
    config_remove_profile,
    config_set_profile,
    config_show,
)
from croud.login import login
from croud.logout import logout
from croud.me import me, me_edit
from croud.organizations.auditlogs.commands import auditlogs_list
from croud.organizations.commands import (
    organizations_create,
    organizations_delete,
    organizations_edit,
    organizations_get,
    organizations_list,
)
from croud.organizations.users.commands import (
    org_users_add,
    org_users_list,
    org_users_remove,
)
from croud.parser import Argument, create_parser
from croud.printer import print_error, print_info
from croud.products.commands import products_list
from croud.projects.commands import (
    project_create,
    project_delete,
    project_edit,
    projects_get,
    projects_list,
)
from croud.projects.users.commands import (
    project_users_add,
    project_users_list,
    project_users_remove,
)
from croud.regions.commands import regions_create, regions_delete, regions_list
from croud.subscriptions.commands import subscriptions_get, subscriptions_list
from croud.tools.spinner import HALO
from croud.users.commands import users_list
from croud.users.roles.commands import roles_list

# fmt: off
command_tree = {
    "me": {
        "help": "Print information about the current logged in user.",
        "resolver": me,
        "commands": {
            "edit": {
                "help": "Edit your user data.",
                "extra_args": [
                    Argument(
                        "--email", type=str, required=True,
                        help="The new email address to use."
                    ),
                ],
                "resolver": me_edit,
            },
        }
    },
    "login": {
        "help": "Log in to CrateDB Cloud.",
        "resolver": login,
        "extra_args": [
            Argument(
                "--idp", type=str, required=False,
                choices={"cognito", "azuread"},
                help="The identity provider (IdP) for the login."
            ),
        ],
    },
    "logout": {"help": "Log out of CrateDB Cloud.", "resolver": logout},
    "config": {
        "help": "Manage croud configuration.",
        "commands": {
            "show": {
                "help": "Show the full configuration.",
                "resolver": config_show,
                "omit": {"sudo", "region", "format"},
            },
            "profiles": {
                "help": "Manage configuration profiles.",
                "commands": {
                    "current": {
                        "help": "Print the current profile.",
                        "resolver": config_current_profile,
                        "omit": {"sudo", "region"},
                    },
                    "use": {
                        "help": "Set the current profile.",
                        "resolver": config_set_profile,
                        "omit": {"sudo", "region"},
                        "extra_args": [
                            Argument(
                                "profile", type=str,
                                help="The name of the profile that should be used."
                            ),
                        ],
                    },
                    "add": {
                        "help": "Add a new profile.",
                        "resolver": config_add_profile,
                        "omit": {"sudo", "region", "format"},
                        "extra_args": [
                            Argument(
                                "profile", type=str,
                                help="The name of the profile that should be added."
                            ),
                            Argument(
                                "--endpoint", type=str, required=True,
                                help="The API endpoint for the profile."
                            ),
                            Argument(
                                "--format", type=str, required=False,
                                help="The output format for the profile."
                            ),
                            Argument(
                                "--region", type=str, required=False,
                                help="The region for the profile."
                            ),
                        ],
                    },
                    "remove": {
                        "help": "Remove an existing profile.",
                        "resolver": config_remove_profile,
                        "omit": {"sudo", "region", "format"},
                        "extra_args": [
                            Argument(
                                "profile", type=str,
                                help="The name of the profile that should be removed."
                            ),
                        ],
                    },
                },
            },
        },
    },
    "projects": {
        "help": "Manage projects.",
        "commands": {
            "create": {
                "help": "Create a project in the specified organization and region.",
                "extra_args": [
                    Argument(
                        "--name", type=str, required=True,
                        help="The project name to use.",
                    ),
                    Argument(
                        "--org-id", type=str, required=False,
                        help="The organization ID to use.",
                    ),
                    Argument(
                        "--backup-location-type",
                        choices=["s3"],
                        type=str,
                        required=False,
                        help="The type of the custom backup location to use. "
                             "Only 's3' currently supported. "
                             "CrateDB Edge regions only."
                    ),
                    Argument(
                        "--backup-location", type=str, required=False,
                        help="The location of where backups are to be stored, "
                             "i.e. name of s3 bucket for s3 locations. "
                             "CrateDB Edge regions only."
                    ),
                    Argument(
                        "--backup-location-access-key-id", type=str, required=False,
                        help="The AWS access key id for the given s3 bucket. "
                             "CrateDB Edge regions only."
                    ),
                    Argument(
                        "--backup-location-endpoint-url", type=str, required=False,
                        help="The URL to a S3 compatible endpoint. "
                             "CrateDB Edge regions only."
                    ),
                    Argument(
                        "--backup-location-secret-access-key", type=str, required=False,
                        help="The AWS secret access key for the given s3 bucket. "
                             "CrateDB Edge regions only."
                    ),
                ],
                "resolver": project_create,
            },
            "delete": {
                "help": "Delete the specified project.",
                "extra_args": [
                    Argument(
                        "-p", "--project-id", type=str, required=True,
                        help="The project ID to use.",
                    ),
                    Argument("-y", "--yes", action="store_true", default=False),
                ],
                "resolver": project_delete,
            },
            "edit": {
                "help": "Edit the specified project.",
                "extra_args": [
                    Argument(
                        "-p", "--project-id", type=str, required=True,
                        help="The project ID to use.",
                    ),
                    Argument(
                        "--name", type=str, required=False,
                        help="The new project name to use.",
                    ),
                ],
                "resolver": project_edit,
            },
            "get": {
                "help": (
                    "Get a project by its ID."
                ),
                "extra_args": [
                    Argument(
                        "id", type=str,
                        help="The ID of the project.",
                    ),
                ],
                "resolver": projects_get,
            },
            "list": {
                "help": (
                    "List all projects the current user has access to in "
                    "the specified region."
                ),
                "resolver": projects_list,
            },
            "users": {
                "help": "Manage users in projects.",
                "commands": {
                    "add": {
                        "help": "Add the selected user to a project.",
                        "extra_args": [
                            Argument(
                                "-p", "--project-id", type=str, required=True,
                                help="The project ID to use.",
                            ),
                            Argument(
                                "--user", type=str, required=True,
                                help="The user email address or ID to use.",
                            ),
                            Argument(
                                "--role", type=str, required=True,
                                help=(
                                    "The role FQN to use. Run `croud users roles list` "
                                    "for a list of available roles."
                                ),
                            ),
                        ],
                        "resolver": project_users_add,
                    },
                    "list": {
                        "help": "List all users within a project.",
                        "extra_args": [
                            Argument(
                                "-p", "--project-id", type=str, required=True,
                                help="The project ID to use.",
                            ),
                        ],
                        "resolver": project_users_list,
                    },
                    "remove": {
                        "help": "Remove the selected user from a project.",
                        "extra_args": [
                            Argument(
                                "-p", "--project-id", type=str, required=True,
                                help="The project ID to use.",
                            ),
                            Argument(
                                "--user", type=str, required=True,
                                help="The user email address or ID to use.",
                            ),
                        ],
                        "resolver": project_users_remove,
                    },
                },
            },
        },
    },
    "clusters": {
        "help": "Manage clusters.",
        "commands": {
            "get": {
                "help": (
                    "Get a cluster by its ID."
                ),
                "extra_args": [
                    Argument(
                        "id", type=str,
                        help="The ID of the cluster.",
                    ),
                ],
                "resolver": clusters_get,
            },
            "list": {
                "help": "List all clusters the current user has access to.",
                "extra_args": [
                    Argument(
                        "-p", "--project-id", type=str, required=False,
                        help="The project ID to use.",
                    ),
                ],
                "resolver": clusters_list,
            },
            "deploy": {
                "help": "Deploy a new CrateDB cluster.",
                "extra_args": [
                    Argument(
                        "--product-name", type=str, required=True,
                        help="The product name to use.",
                    ),
                    Argument(
                        "--tier", type=str, required=True,
                        help="The product tier to use.",
                    ),
                    Argument(
                        "--unit", type=int, required=False,
                        help="The product scale unit to use.",
                    ),
                    Argument(
                        "-p", "--project-id", type=str, required=True,
                        help="The project ID to use.",
                    ),
                    Argument(
                        "--cluster-name", type=str, required=True,
                        help="The CrateDB cluster name to use.",
                    ),
                    Argument(
                        "--version", type=str, required=True,
                        help="The CrateDB version to use.",
                    ),
                    Argument(
                        "--username", type=str, required=True,
                        help="The CrateDB username to use.",
                    ),
                    Argument(
                        "--password", type=str, required=True,
                        help="The CrateDB password to use.",
                    ),
                    Argument(
                        "--channel", type=str, default="stable", required=False,
                        choices={"stable", "testing", "nightly"},
                        help="The channel of the CrateDB version (superusers only).",
                    ),
                    Argument(
                        "--cpus", type=float, required=False,
                        help="Number of CPU cores to allocate. Can be fractional. "
                             "CrateDB Edge regions only.",
                    ),
                    Argument(
                        "--disks", type=int, required=False,
                        help="Number of disks to attach. "
                             "CrateDB Edge regions only.",
                    ),
                    Argument(
                        "--disk-size-gb", type=int, required=False,
                        help="Size of disks to attach (in GiB). "
                             "CrateDB Edge regions only.",
                    ),
                    Argument(
                        "--disk-type", type=str, required=False,
                        choices={"standard", "premium"},
                        help="Type of disks to use. "
                             "CrateDB Edge regions only.",
                    ),
                    Argument(
                        "--memory-size-mb", type=int, required=False,
                        help="Amount of memory to allocate (in MiB). "
                             "CrateDB Edge regions only.",
                    )
                ],
                "resolver": clusters_deploy,
            },
            "scale": {
                "help": "Scale an existing CrateDB cluster.",
                "extra_args": [
                    Argument(
                        "--cluster-id", type=str, required=True,
                        help="The CrateDB cluster ID to use.",
                    ),
                    Argument(
                        "--unit", type=int, required=True,
                        help="The product scale unit to use.",
                    ),
                ],
                "resolver": clusters_scale,
            },
            "upgrade": {
                "help": "Upgrade an existing CrateDB cluster to a later version.",
                "extra_args": [
                    Argument(
                        "--cluster-id", type=str, required=True,
                        help="The CrateDB cluster ID to use.",
                    ),
                    Argument(
                        "--version", type=str, required=True,
                        help="The CrateDB version to use.",
                    ),
                ],
                "resolver": clusters_upgrade
            },
            "delete": {
                "help": "Delete the specified cluster.",
                "extra_args": [
                    Argument(
                        "--cluster-id", type=str, required=True,
                        help="The CrateDB cluster ID to use.",
                    ),
                    Argument("-y", "--yes", action="store_true", default=False)
                ],
                "resolver": clusters_delete,
            },
            "restart-node": {
                "help": "Restart a node in a CrateDB cluster.",
                "extra_args": [
                    Argument(
                        "--cluster-id", type=str, required=True,
                        help="The CrateDB cluster ID to use.",
                    ),
                    Argument(
                        "--ordinal", type=int, required=True,
                        help="The ordinal index of the node to restart.",
                    ),
                ],
                "resolver": clusters_restart_node,
            },
            "set-deletion-protection": {
                "help": "Set the deletion protection status of a CrateDB cluster.",
                "extra_args": [
                    Argument(
                        "--cluster-id", type=str, required=True,
                        help="The CrateDB cluster ID to use.",
                    ),
                    Argument(
                        "--value", type=lambda x: bool(strtobool(str(x))),
                        required=True, help="The deletion protection status",
                    ),
                ],
                "resolver": clusters_set_deletion_protection,
            },
        },
    },
    "products": {
        "help": "Manage Products.",
        "commands": {
            "list": {
                "help": "List all available products in the current region.",
                "extra_args": [
                    Argument(
                        "--kind", type=str, required=False, help="The product kind."
                    ),
                ],
                "resolver": products_list,
            },
        },
    },
    "organizations": {
        "help": "Manage organizations.",
        "commands": {
            "create": {
                "help": "Create a new organization.",
                "extra_args": [
                    Argument(
                        "--name", type=str, required=True,
                        help="The organization name to use.",
                    ),
                    Argument(
                        "--plan-type", type=int, required=False,
                        choices=[1, 2, 3, 4, 5, 6],
                        help="The support plan to use for the organization. Argument "
                             "is for superusers only.",
                    ),
                ],
                "resolver": organizations_create,
            },
            "get": {
                "help": (
                    "Get an organization by its ID."
                ),
                "extra_args": [
                    Argument(
                        "id", type=str,
                        help="The ID of the organization.",
                    ),
                ],
                "resolver": organizations_get,
            },
            "list": {
                "help": "List all organizations the current user has access to.",
                "resolver": organizations_list,
            },
            "edit": {
                "help": "Edit the specified organization.",
                "extra_args": [
                    Argument(
                        "--name", type=str, required=False,
                        help="The new organization name to use.",
                    ),
                    Argument(
                        "--plan-type", type=int, required=False,
                        choices=[1, 2, 3, 4, 5, 6],
                        help="The new support plan to use for the organization.",
                    ),
                    Argument(
                        "--org-id", type=str, required=False,
                        help="The organization ID to use.",
                    ),
                ],
                "resolver": organizations_edit,
            },
            "delete": {
                "help": "Delete the specified organization.",
                "extra_args": [
                    Argument("-y", "--yes", action="store_true", default=False),
                    Argument(
                        "--org-id", type=str, required=False,
                        help="The organization ID to use.",
                    ),
                ],
                "resolver": organizations_delete,
            },
            "auditlogs": {
                "help": "Show audit logs for an organization.",
                "commands": {
                    "list": {
                        "help": "List all audit events in the current organization.",
                        "extra_args": [
                            Argument(
                                "--action", type=str, required=False,
                                help="The audit event action.",
                            ),
                            Argument(
                                "--from", type=str, required=False, dest="from_",
                                help="Only show events from this point in time.",
                                metavar="FROM",
                            ),
                            Argument(
                                "--to", type=str, required=False,
                                help="Only show events older than this.",
                            ),
                            Argument(
                                "--org-id", type=str, required=False,
                                help="The organization ID to use.",
                            ),
                        ],
                        "resolver": auditlogs_list,
                    },
                },
            },
            "users": {
                "help": "Manage users in an organization.",
                "commands": {
                    "add": {
                        "help": "Add the selected user to an organization.",
                        "extra_args": [
                            Argument(
                                "--user", type=str, required=True,
                                help="The user email address or ID to use.",
                            ),
                            Argument(
                                "--role", type=str, required=True,
                                help=(
                                    "The role FQN to use. Run `croud users roles list` "
                                    "for a list of available roles."
                                ),
                            ),
                            Argument(
                                "--org-id", type=str, required=False,
                                help="The organization ID to use.",
                            ),
                        ],
                        "resolver": org_users_add,
                    },
                    "list": {
                        "help": "List all users within an organization.",
                        "extra_args": [
                            Argument(
                                "--org-id", type=str, required=False,
                                help="The organization ID to use.",
                            ),
                        ],
                        "resolver": org_users_list,
                    },
                    "remove": {
                        "help": "Remove the selected user from an organization.",
                        "extra_args": [
                            Argument(
                                "--user", type=str, required=True,
                                help="The user email address or ID to use.",
                            ),
                            Argument(
                                "--org-id", type=str, required=False,
                                help="The organization ID to use.",
                            ),
                        ],
                        "resolver": org_users_remove,
                    },
                },
            },
        },
    },
    "users": {
        "help": "Manage users.",
        "commands": {
            "list": {
                "help": "List all users.",
                "extra_args": [
                    Argument(
                        "--no-roles", action="store_true",
                        help="List users without roles.",
                    ),
                    Argument(
                        "--no-org", action="store_true",
                        help="Only show users that are not part of any organization.",
                    )
                ],
                "resolver": users_list,
            },
            "roles": {
                "help": "Manage user roles.",
                "commands": {
                    "list": {
                        "help": "List all available roles.",
                        "resolver": roles_list,
                    },
                },
            },
        },
    },
    "regions": {
        "help": "Print information about available regions.",
        "commands": {
            "list": {
                "help": "List all available regions.",
                "resolver": regions_list,
            },
            "create": {
                "help": "Create a new region.",
                "resolver": regions_create,
                "extra_args": [
                    Argument(
                        "--aws-bucket", type=str, required=False,
                        help="The AWS S3 bucket where cluster backups will be stored.",
                    ),
                    Argument(
                        "--aws-region", type=str, required=False,
                        help="The AWS region where the S3 bucket for cluster"
                             " backups is expected to be.",
                    ),
                    Argument(
                        "--name", type=str, required=False,
                        help="The domain name prefix that will be used to reach"
                             " the region. A Crate domain will be added as a suffix.",
                    ),
                    Argument(
                        "--description", type=str, required=True,
                        help="The description of the new region.",
                    ),
                    Argument(
                        "--provider", type=str, required=True,
                        help="Either AWS, EKS or EDGE",
                    ),
                    Argument(
                        "--org-id", type=str, required=False,
                        help="The organization ID to use.",
                    ),
                    Argument("-y", "--yes", action="store_true", default=False),
                ],
            },
            "delete": {
                "help": "Delete an existing edge region.",
                "resolver": regions_delete,
                "extra_args": [
                    Argument(
                        "--name", type=str, required=True,
                        help="The name of the region that will be deleted."
                    ),
                    Argument("-y", "--yes", action="store_true", default=False),
                ],
            },
        }
    },
    "subscriptions": {
        "help": "Manage subscriptions.",
        "commands": {
            "get": {
                "help": (
                    "Get a subscription by its ID."
                ),
                "extra_args": [
                    Argument(
                        "id", type=str,
                        help="The ID of the subscription.",
                    ),
                ],
                "resolver": subscriptions_get,
            },
            "list": {
                "help": "List all subscriptions the current user has access to.",
                "resolver": subscriptions_list,
            },
        },
    },
}
# fmt: on


def get_parser():
    tree = {
        "help": "A command line interface for CrateDB Cloud.",
        "commands": command_tree,
    }
    return create_parser(tree)


def main():
    if not CONFIG.is_valid():
        print_error(
            "Your configuration file is incompatible with the current version of croud."
        )
        print_info(
            f"Please delete the file '{CONFIG._file_path}' or update it manually."
        )
        sys.exit(1)

    colorama.init()

    parser = get_parser()
    params = parser.parse_args(sys.argv[1:])
    if "resolver" in params:
        fn = params.resolver
        del params.resolver
        with HALO:
            fn(params)
    else:
        parser.print_help()


if __name__ == "__main__":
    main()



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

import re
from typing import Any, Dict

SENSITIVE_KEYS = re.compile("pass|secret|token", flags=re.IGNORECASE)


def clean_dict(data: Dict[str, Any]) -> Dict[str, Any]:
    """
    Clean "secrets" from a dict.

    This is solely used for masking the "auth-token" value of the configuration
    dictionary when printing it to the console.

    It can only handle "pure" dicts with strings as keys and dicts, lists and
    simple types as values.
    """

    def clean(k, v):
        if SENSITIVE_KEYS.search(k):
            return "x" * 10
        elif isinstance(v, list):
            cleaned_list = []
            for item in v:
                if isinstance(item, dict):
                    cleaned_list.append(clean_dict(item))
                else:
                    cleaned_list.append(item)
            return cleaned_list
        elif isinstance(v, dict):
            return clean_dict(v)
        return v

    return {k: clean(k, v) for k, v in data.items()}



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

import re
from argparse import Namespace
from typing import Dict, List

from croud.api import Client
from croud.config import get_output_format
from croud.printer import print_error, print_response
from croud.util import org_id_config_fallback

# Hat tip to Django for ISO8601 deserialization functions

iso8601_datetime_re = re.compile(
    r"(?P<year>\d{4})-(?P<month>\d{1,2})-(?P<day>\d{1,2})"
    r"[T ](?P<hour>\d{1,2}):(?P<minute>\d{1,2})"
    r"(?::(?P<second>\d{1,2})(?:\.(?P<microsecond>\d{1,6})\d{0,6})?)?"
    r"(?P<tzinfo>Z|[+-]\d{2}(?::?\d{2})?)?$"
)


def actor_id_transform(field):
    return field["id"] or "SYSTEM"


@org_id_config_fallback
def auditlogs_list(args: Namespace) -> None:
    client = Client.from_args(args)
    url = f"/api/v2/organizations/{args.org_id}/auditlogs/"
    data: List[Dict] = []
    cursor = None

    params = {}
    if args.action:
        params["action"] = args.action

    if args.from_:
        if not iso8601_datetime_re.fullmatch(args.from_):
            print_error("Invalid 'from' date format.")
        params["from"] = args.from_

    if args.to:
        if not iso8601_datetime_re.fullmatch(args.to):
            print_error("Invalid 'to' date format.")
        params["to"] = args.to

    while True:
        if cursor:
            params["last"] = cursor
        page, errors = client.get(url, params=params)
        if errors or not page:
            break
        else:
            data.extend(page)
            cursor = data[-1]["id"]

    print_response(
        data=data,
        errors=errors,
        keys=["action", "actor", "created"],
        output_fmt=get_output_format(args),
        transforms={"actor": actor_id_transform},
    )



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

import re
from argparse import Namespace
from functools import partial

from croud.config.configuration import (  # mark for IDEs as "private"
    Configuration as _Configuration,
)

SENSITIVE_KEYS = re.compile("pass|secret|token", flags=re.IGNORECASE)
_CONFIG = None


def _get_config(filename: str) -> _Configuration:
    """
    Lazily instantiate and return the module level ``_CONFIG`` object.

    This method is used as a getter, to make sure there is only a single
    instance of the globally used configuration.
    """
    global _CONFIG
    if _CONFIG is None:
        _CONFIG = _Configuration(filename)
    return _CONFIG


class WeakConfigProxy:
    """
    A proxy class for the global Configuration instance.

    By providing a getter method for the proxied object, the proxied object may
    be replaced without the proxy knowing it, due to the nature of the weak
    coupling.
    """

    def __init__(self, getter, *args, **kwargs):
        super().__setattr__("_get", partial(getter, *args, **kwargs))

    def __repr__(self):
        return repr(self._get())

    def __getattr__(self, name):
        return getattr(self._get(), name)

    def __setattr__(self, name, value):
        raise AttributeError("Cannot set attributes on proxy object")


# The global Configuration object
CONFIG = WeakConfigProxy(_get_config, "croud.yaml")


# This method should rather be in the `croud.config.util` module,
# however, because it depends on `CONFIG`, an import of `CONFIG`
# inside the `util` module would sooner or later result in a circular
# import.
def get_output_format(args: Namespace) -> str:
    """
    Get the output format for the response.

    The fallback chain is as follows:

    1. --output-fmt FORMAT
    2. format of current profile
    3. default format
    """
    return args.output_fmt if args.output_fmt is not None else CONFIG.format



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

from unittest import mock

import pytest

from croud.login import get_org_id, login_path
from croud.server import Server
from tests.util import call_command


@mock.patch.object(Server, "wait_for_shutdown")
@mock.patch.object(Server, "start_in_background")
@mock.patch("croud.login.can_launch_browser", return_value=True)
@mock.patch("croud.login.open_page_in_browser")
@mock.patch("croud.login.print_info")
def test_login(
    mock_print_info,
    mock_open_page_in_browser,
    mock_can_launch_browser,
    mock_start_in_background,
    mock_wait_for_shutdown,
    config,
):
    with mock.patch("croud.login.get_org_id", return_value="my-org-id"):
        call_command("croud", "login")

    calls = [
        mock.call("A browser tab has been launched for you to login."),
        mock.call("Login successful."),
    ]
    mock_print_info.assert_has_calls(calls)
    assert config.profile["organization-id"] == "my-org-id"


@mock.patch("croud.login.can_launch_browser", return_value=False)
@mock.patch("croud.login.print_error")
def test_login_no_valid_browser(mock_print_error, mock_can_launch_browser, config):
    with pytest.raises(SystemExit) as e_info:
        call_command("croud", "login")

    mock_print_error.assert_called_once_with(
        "Login only works with a valid browser installed."
    )
    assert e_info.value.code == 1


@pytest.mark.parametrize("org_id_param", [None, "some-id"])
def test_get_org_id(org_id_param, config):
    with mock.patch(
        "croud.api.Client.request",
        return_value=[{"organization_id": org_id_param}, None],
    ):
        org_id = get_org_id()
    assert org_id == org_id_param


@pytest.mark.parametrize(
    "idp,expected",
    [(None, "/oauth2/login?cli=true"), ("idp", "/oauth2/idp/login?cli=true")],
)
def test_login_path(idp, expected):
    assert login_path(idp) == expected



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

import uuid
from unittest import mock

import pytest

from croud.api import Client, RequestMethod
from croud.projects.users.commands import (
    role_fqn_transform as project_role_fqn_transform,
)
from tests.util import assert_rest, call_command, gen_uuid

pytestmark = pytest.mark.usefixtures("config")


@mock.patch.object(Client, "request", return_value=({}, None))
def test_projects_create(mock_request):
    call_command(
        "croud",
        "projects",
        "create",
        "--name",
        "new-project",
        "--org-id",
        "organization-id",
    )
    assert_rest(
        mock_request,
        RequestMethod.POST,
        "/api/v2/projects/",
        body={"name": "new-project", "organization_id": "organization-id"},
    )


@mock.patch.object(Client, "request", return_value=({}, None))
def test_projects_create_with_custom_backup(mock_request):
    call_command(
        "croud",
        "projects",
        "create",
        "--name",
        "new-project",
        "--org-id",
        "organization-id",
        "--backup-location",
        "some",
        "--backup-location-type",
        "s3",
        "--backup-location-access-key-id",
        "123",
        "--backup-location-secret-access-key",
        "321",
        "--backup-location-endpoint-url",
        "https://minio.svc.local",
    )
    assert_rest(
        mock_request,
        RequestMethod.POST,
        "/api/v2/projects/",
        body={
            "name": "new-project",
            "organization_id": "organization-id",
            "backup_location": {
                "location_type": "s3",
                "location": "some",
                "credentials": {"access_key_id": "123", "secret_access_key": "321"},
                "additional_config": {"endpoint_url": "https://minio.svc.local"},
            },
        },
    )


@mock.patch.object(Client, "request", return_value=(None, {}))
def test_projects_delete(mock_request, capsys):
    project_id = gen_uuid()
    with mock.patch("builtins.input", side_effect=["yes"]) as mock_input:
        call_command("croud", "projects", "delete", "--project-id", project_id)
    assert_rest(mock_request, RequestMethod.DELETE, f"/api/v2/projects/{project_id}/")
    mock_input.assert_called_once_with(
        "Are you sure you want to delete the project? [yN] "
    )

    _, err_output = capsys.readouterr()
    assert "Success" in err_output
    assert "Project deleted." in err_output


@mock.patch.object(Client, "request", return_value=(None, {}))
def test_projects_delete_flag(mock_request, capsys):
    project_id = gen_uuid()
    with mock.patch("builtins.input", side_effect=["y"]) as mock_input:
        call_command("croud", "projects", "delete", "--project-id", project_id, "-y")
    assert_rest(mock_request, RequestMethod.DELETE, f"/api/v2/projects/{project_id}/")
    mock_input.assert_not_called()

    _, err_output = capsys.readouterr()
    assert "Success" in err_output
    assert "Project deleted." in err_output


@mock.patch.object(Client, "request", return_value=(None, {}))
def test_projects_delete_aborted(mock_request, capsys):
    project_id = gen_uuid()
    with mock.patch("builtins.input", side_effect=["Nooooo"]) as mock_input:
        call_command("croud", "projects", "delete", "--project-id", project_id)
    mock_request.assert_not_called()
    mock_input.assert_called_once_with(
        "Are you sure you want to delete the project? [yN] "
    )

    _, err_output = capsys.readouterr()
    assert "Project deletion cancelled." in err_output


@mock.patch.object(
    Client,
    "request",
    return_value=(
        [
            {
                "id": "1",
                "name": "Test Project",
                "organization_id": "test.org.id",
                "region": "aks1.eastus.azure",
                "backup_location": {"location_type": "s3", "location": "some"},
            },
            {
                "id": "2",
                "name": "Test Project 2",
                "organization_id": "test.org.id",
                "region": "aks1.eastus.azure",
                "backup_location": None,
            },
        ],
        None,
    ),
)
def test_projects_list(mock_request):
    call_command("croud", "projects", "list")
    assert_rest(mock_request, RequestMethod.GET, "/api/v2/projects/")


@mock.patch.object(Client, "request", return_value=({}, None))
def test_projects_get(mock_request):
    id = str(uuid.uuid4())
    call_command("croud", "projects", "get", id)
    assert_rest(mock_request, RequestMethod.GET, f"/api/v2/projects/{id}/")


@pytest.mark.parametrize(
    "added,message",
    [(True, "User added to project."), (False, "Role altered for user.")],
)
@mock.patch.object(Client, "request")
def test_projects_users_add(mock_request, added, message, capsys):
    mock_request.return_value = ({"added": added}, None)

    project_id = gen_uuid()

    # uid or email would be possible for the backend
    user = "test@crate.io"
    role_fqn = "project_admin"

    call_command(
        "croud",
        "projects",
        "users",
        "add",
        "--project-id",
        project_id,
        "--user",
        user,
        "--role",
        role_fqn,
    )
    assert_rest(
        mock_request,
        RequestMethod.POST,
        f"/api/v2/projects/{project_id}/users/",
        body={"user": user, "role_fqn": role_fqn},
    )

    _, err_output = capsys.readouterr()
    assert "Success" in err_output
    assert message in err_output


@mock.patch.object(Client, "request", return_value=({}, None))
def test_projects_users_list(mock_request):
    project_id = gen_uuid()

    call_command("croud", "projects", "users", "list", "--project-id", project_id)
    assert_rest(
        mock_request, RequestMethod.GET, f"/api/v2/projects/{project_id}/users/"
    )


@mock.patch.object(Client, "request", return_value=({}, None))
def test_projects_users_remove(mock_request):
    project_id = gen_uuid()

    # uid or email would be possible for the backend
    user = "test@crate.io"

    call_command(
        "croud",
        "projects",
        "users",
        "remove",
        "--project-id",
        project_id,
        "--user",
        user,
    )
    assert_rest(
        mock_request,
        RequestMethod.DELETE,
        f"/api/v2/projects/{project_id}/users/{user}/",
    )


def test_role_fqn_transform():
    user = {
        "project_roles": [
            {"project_id": "project-1", "role_fqn": "project_admin"},
            {"project_id": "project-2", "role_fqn": "project_member"},
            {"project_id": "project-3", "role_fqn": "project_member"},
        ]
    }
    response = project_role_fqn_transform(user["project_roles"])
    assert response == "project_admin"


@mock.patch.object(Client, "request", return_value=(None, {}))
def test_projects_edit(mock_request, capsys):
    project_id = gen_uuid()
    call_command("croud", "projects", "edit", "-p", project_id, "--name", "new-name")
    assert_rest(
        mock_request,
        RequestMethod.PATCH,
        f"/api/v2/projects/{project_id}/",
        body={"name": "new-name"},
    )

    _, err_output = capsys.readouterr()
    assert "Success" in err_output
    assert "Project edited" in err_output


@mock.patch.object(Client, "request", return_value=(None, {}))
def test_projects_edit_no_argument(mock_request, capsys):
    project_id = gen_uuid()
    with pytest.raises(SystemExit):
        call_command("croud", "projects", "edit", "-p", project_id)

    _, err_output = capsys.readouterr()
    assert "Error" in err_output
    assert "No input arguments found." in err_output



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

from http.server import BaseHTTPRequestHandler, HTTPServer
from threading import Thread
from typing import Callable
from urllib import parse

HOST = "localhost"
PORT = 8400


class SetTokenHTTPServer(HTTPServer):
    def __init__(self, on_token: Callable[[str], None], *args, **kwargs):
        self.on_token = on_token
        super().__init__(*args, **kwargs)


class SetTokenHandler(BaseHTTPRequestHandler):
    SUCCESS_MSG = (
        b"<b>You have successfully logged into CrateDB Cloud!</b>"
        b"<p>This window can be closed.</p>"
    )
    MISSING_TOKEN_MSG = (
        b"<b>Login to CrateDB Cloud failed!</b>"
        b"<p>Authentication token missing in URL.</p>"
    )
    DUPLICATE_TOKEN_MSG = (
        b"<b>Login to CrateDB Cloud failed!</b>"
        b"<p>More than one authentication token present in URL.</p>"
    )

    def do_GET(self):
        query_string = parse.urlparse(self.path).query
        query = parse.parse_qs(query_string)
        if "token" not in query:
            code = 400
            msg = SetTokenHandler.MISSING_TOKEN_MSG
        elif len(query["token"]) != 1:
            code = 400
            msg = SetTokenHandler.DUPLICATE_TOKEN_MSG
        else:
            token = query["token"][0]
            self.server.on_token(token)
            code = 200
            msg = SetTokenHandler.SUCCESS_MSG
        self.send_response(code)
        self.send_header("Content-Type", "text/html")
        self.send_header("Content-Length", str(len(msg)))
        self.end_headers()
        self.wfile.write(msg)
        # Set the termination condition for the server's `serve_forever`.
        # We cannot use `self.server.shutdown()` here because we're in the same
        # thread as the one who'd be waiting for the shutdown, thus causing a
        # deadlock.
        self.server._BaseServer__shutdown_request = True

    def log_request(self, *args, **kwargs):
        # Just don't log anything ...
        pass


class Server:
    def __init__(self, on_token, random_port: bool = False):
        port = 0 if random_port else PORT
        self._server = SetTokenHTTPServer(on_token, (HOST, port), SetTokenHandler)
        self._thread = Thread(target=self._server.serve_forever, daemon=True)

    def start_in_background(self) -> "Server":
        self._thread.start()
        return self

    def wait_for_shutdown(self, timeout=None):
        self._thread.join(timeout=timeout)

    @property
    def port(self):
        return self._server.socket.getsockname()[1]



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

import argparse
import re
from platform import python_version

import pytest

import croud
from croud.api import Client


def test_send_success_sets_data_with_key(client: Client):
    resp_data, errors = client.get("/data/data-key")
    assert resp_data == {"data": {"key": "value"}}
    assert errors is None


def test_send_success_sets_data_without_key(client: Client):
    resp_data, errors = client.get("/data/no-key")
    assert resp_data == {"key": "value"}
    assert errors is None


def test_send_error_sets_error(client: Client):
    resp_data, errors = client.get("/errors/400")
    assert errors == {"message": "Bad request.", "errors": {"key": "Error on 'key'"}}


def test_send_text_response(client: Client):
    resp_data, errors = client.get("/text-response")
    assert resp_data is None
    assert errors == {"message": "500 - Internal Server Error", "success": False}


def test_send_empty_response(client: Client):
    resp_data, errors = client.get("/empty-response")
    assert resp_data is None
    assert errors is None


def test_send_redirect_response(client: Client, capsys):
    with pytest.raises(SystemExit):
        client.get("/redirect")
    _, err = capsys.readouterr()
    assert "Unauthorized." in err
    assert "Use `croud login` to login to CrateDB Cloud." in err


def test_send_new_token_response(client: Client, config):
    client.get("/new-token")
    assert ("session", "new-token") in client.session.cookies.items()
    assert config.token == client._token == "new-token"


@pytest.mark.parametrize("sudo", [True, False])
def test_send_sudo_header(sudo, config):
    client = Client(config.endpoint, region="bregenz.a1", sudo=sudo, _verify_ssl=False)
    resp_data, errors = client.get("/client-headers")
    assert ("X-Auth-Sudo" in resp_data) == sudo  # type: ignore


def test_send_user_agent_header(config):
    client = Client(config.endpoint, region="bregenz.a1", _verify_ssl=False)
    resp_data, errors = client.get("/client-headers")
    expected_ua = f"Croud/{croud.__version__} Python/{python_version()}"
    assert resp_data["User-Agent"] == expected_ua


@pytest.mark.parametrize(
    "argument,is_header_present", [(None, False), ("westeurope.azure", True)]
)
def test_send_region_header(argument, is_header_present, config):
    client = Client(config.endpoint, region=argument, _verify_ssl=False)
    resp_data, errors = client.get("/client-headers")
    if is_header_present:
        assert resp_data["X-Region"] == argument  # type: ignore
    else:
        assert "X-Region" not in resp_data


def test_client_initialization(config):
    args = argparse.Namespace(sudo=True, region="bregenz.a1")
    client = Client.from_args(args)
    assert str(client.base_url) == config.endpoint
    assert client.session.headers["X-Region"] == "bregenz.a1"
    assert client.session.headers["X-Auth-Sudo"] == "1"
    assert client.session.cookies["session"] == config.token


def test_error_message_on_connection_error(config):
    config.add_profile("test", endpoint="https://invalid.cratedb.local")
    config.set_auth_token("test", "some-token")
    config.use_profile("test")

    expected_message_re = re.compile(
        r"^Failed to perform request on 'https://invalid\.cratedb\.local/me\'. "
        r"Original error was: '.*'$"
    )
    client = Client(config.endpoint, _verify_ssl=False)
    resp_data, errors = client.get("/me")
    print(resp_data, errors)
    assert resp_data is None
    assert expected_message_re.match(errors["message"]) is not None
    assert errors["success"] is False



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

import uuid
from unittest import mock

import pytest

from croud.api import Client, RequestMethod
from croud.organizations.users.commands import (
    role_fqn_transform as organization_role_fqn_transform,
)
from tests.util import assert_rest, call_command, gen_uuid

pytestmark = pytest.mark.usefixtures("config")

config_org_id = gen_uuid()
FALLBACK_ORG_ID_CONFIG: dict = {
    "auth": {
        "current_context": "prod",
        "contexts": {
            "prod": {"token": "", "organization_id": config_org_id},
            "dev": {"token": "", "organization_id": config_org_id},
            "local": {"token": "", "organization_id": config_org_id},
        },
    },
    "region": "bregenz.a1",
    "output_fmt": "table",
}


@mock.patch.object(Client, "request", return_value=({}, None))
def test_organizations_create(mock_request):
    call_command("croud", "organizations", "create", "--name", "test-org")
    assert_rest(
        mock_request,
        RequestMethod.POST,
        "/api/v2/organizations/",
        body={"name": "test-org"},
    )


@mock.patch.object(Client, "request", return_value=({}, None))
def test_organizations_create_plan_type(mock_request):
    call_command(
        "croud", "organizations", "create", "--name", "test-org", "--plan-type", "3"
    )
    assert_rest(
        mock_request,
        RequestMethod.POST,
        "/api/v2/organizations/",
        body={"name": "test-org", "plan_type": 3},
    )


@mock.patch.object(Client, "request", return_value=({}, None))
def test_organizations_edit(mock_request):
    org_id = gen_uuid()
    call_command(
        "croud",
        "organizations",
        "edit",
        "--name",
        "new-org-name",
        "--plan-type",
        "3",
        "--org-id",
        org_id,
    )
    assert_rest(
        mock_request,
        RequestMethod.PUT,
        f"/api/v2/organizations/{org_id}/",
        body={"name": "new-org-name", "plan_type": 3},
    )


@mock.patch.object(Client, "request", return_value=({}, None))
def test_organizations_edit_name(mock_request):
    org_id = gen_uuid()
    call_command(
        "croud", "organizations", "edit", "--name", "new-org-name", "--org-id", org_id
    )
    assert_rest(
        mock_request,
        RequestMethod.PUT,
        f"/api/v2/organizations/{org_id}/",
        body={"name": "new-org-name"},
    )


@mock.patch.object(Client, "request", return_value=({}, None))
def test_organizations_edit_plan_type(mock_request):
    org_id = gen_uuid()
    call_command(
        "croud", "organizations", "edit", "--plan-type", "3", "--org-id", org_id
    )
    assert_rest(
        mock_request,
        RequestMethod.PUT,
        f"/api/v2/organizations/{org_id}/",
        body={"plan_type": 3},
    )


@mock.patch.object(Client, "request", return_value=(None, {}))
def test_organizations_edit_no_arguments(mock_request, capsys):
    org_id = gen_uuid()
    with pytest.raises(SystemExit):
        call_command("croud", "organizations", "edit", "--org-id", org_id)

    _, err = capsys.readouterr()
    assert "No input arguments found." in err


@mock.patch.object(Client, "request", return_value=({}, None))
def test_organizations_list(mock_request):
    call_command("croud", "organizations", "list")
    assert_rest(mock_request, RequestMethod.GET, "/api/v2/organizations/")


@mock.patch.object(Client, "request", return_value=({}, None))
def test_organizations_get(mock_request):
    id = str(uuid.uuid4())
    call_command("croud", "organizations", "get", id)
    assert_rest(mock_request, RequestMethod.GET, f"/api/v2/organizations/{id}/")


@mock.patch.object(Client, "request", return_value=(None, {}))
def test_organizations_delete(mock_request, capsys):
    org_id = gen_uuid()
    with mock.patch("builtins.input", side_effect=["Y"]) as mock_input:
        call_command("croud", "organizations", "delete", "--org-id", org_id)
    assert_rest(mock_request, RequestMethod.DELETE, f"/api/v2/organizations/{org_id}/")
    mock_input.assert_called_once_with(
        "Are you sure you want to delete the organization? [yN] "
    )

    _, err_output = capsys.readouterr()
    assert "Organization deleted." in err_output


@mock.patch.object(Client, "request", return_value=(None, {"errors": "Message"}))
def test_organizations_delete_failure_org_id_not_deleted_from_config(
    mock_request, capsys, config
):
    org_id = gen_uuid()
    config.set_organization_id(config.name, org_id)

    with mock.patch("builtins.input", side_effect=["Y"]):
        call_command("croud", "organizations", "delete")

    assert_rest(
        mock_request, RequestMethod.DELETE, f"/api/v2/organizations/{org_id}/",
    )

    _, err = capsys.readouterr()
    assert err is not None
    assert config.organization == org_id


@mock.patch.object(Client, "request", return_value=({}, None))
def test_organizations_delete_org_id_from_local_config(
    mock_request, capsys, config,
):
    org_id = gen_uuid()
    config.set_organization_id(config.name, org_id)

    with mock.patch("builtins.input", side_effect=["Y"]):
        call_command("croud", "organizations", "delete")

    assert_rest(
        mock_request, RequestMethod.DELETE, f"/api/v2/organizations/{org_id}/",
    )

    assert config.organization is None


@mock.patch.object(Client, "request", return_value=(None, {}))
def test_organizations_delete_flag(mock_request, capsys):
    org_id = gen_uuid()
    with mock.patch("builtins.input") as mock_input:
        call_command("croud", "organizations", "delete", "--org-id", org_id, "-y")
    assert_rest(mock_request, RequestMethod.DELETE, f"/api/v2/organizations/{org_id}/")
    mock_input.assert_not_called()

    _, err_output = capsys.readouterr()
    assert "Organization deleted." in err_output


@pytest.mark.parametrize("input", ["", "N", "No", "cancel"])
@mock.patch.object(Client, "request", return_value=(None, {}))
def test_organizations_delete_aborted(mock_request, capsys, input):
    org_id = gen_uuid()
    with mock.patch("builtins.input", side_effect=[input]) as mock_input:
        call_command("croud", "organizations", "delete", "--org-id", org_id)
    mock_request.assert_not_called()
    mock_input.assert_called_once_with(
        "Are you sure you want to delete the organization? [yN] "
    )

    _, err_output = capsys.readouterr()
    assert "Organization deletion cancelled." in err_output


@mock.patch.object(Client, "request", return_value=(None, {}))
def test_organizations_delete_aborted_with_input(mock_request, capsys):
    org_id = gen_uuid()
    with mock.patch("builtins.input", side_effect=["N"]) as mock_input:
        call_command("croud", "organizations", "delete", "--org-id", org_id)
    mock_request.assert_not_called()
    mock_input.assert_called_once_with(
        "Are you sure you want to delete the organization? [yN] "
    )

    _, err_output = capsys.readouterr()
    assert "Organization deletion cancelled." in err_output


@mock.patch.object(Client, "request", return_value=({}, None))
def test_organizations_auditlogs_list(mock_request):
    org_id = gen_uuid()

    call_command("croud", "organizations", "auditlogs", "list", "--org-id", org_id)
    assert_rest(
        mock_request,
        RequestMethod.GET,
        f"/api/v2/organizations/{org_id}/auditlogs/",
        params={},
    )


@mock.patch.object(Client, "request", return_value=({}, None))
def test_organizations_auditlogs_list_filtered(mock_request):
    org_id = gen_uuid()

    call_command(
        "croud",
        "organizations",
        "auditlogs",
        "list",
        "--org-id",
        org_id,
        "--action",
        "organization.create",
        "--from",
        "2019-10-11T12:13:14",
        "--to",
        "2019-11-12T12:34:56",
    )
    assert_rest(
        mock_request,
        RequestMethod.GET,
        f"/api/v2/organizations/{org_id}/auditlogs/",
        params={
            "action": "organization.create",
            "from": "2019-10-11T12:13:14",
            "to": "2019-11-12T12:34:56",
        },
    )


@pytest.mark.parametrize(
    "added,message",
    [(True, "User added to organization."), (False, "Role altered for user.")],
)
@mock.patch.object(Client, "request")
def test_organizations_users_add(mock_request, added, message, capsys):
    mock_request.return_value = ({"added": added}, None)

    org_id = gen_uuid()
    user = "test@crate.io"
    role_fqn = "org_admin"

    call_command(
        "croud",
        "organizations",
        "users",
        "add",
        "--user",
        user,
        "--org-id",
        org_id,
        "--role",
        role_fqn,
    )
    assert_rest(
        mock_request,
        RequestMethod.POST,
        f"/api/v2/organizations/{org_id}/users/",
        body={"user": user, "role_fqn": role_fqn},
    )

    _, err_output = capsys.readouterr()
    assert "Success" in err_output
    assert message in err_output


@mock.patch.object(Client, "request", return_value=({}, None))
def test_organizations_users_list(mock_request):
    org_id = gen_uuid()

    call_command("croud", "organizations", "users", "list", "--org-id", org_id)
    assert_rest(
        mock_request, RequestMethod.GET, f"/api/v2/organizations/{org_id}/users/"
    )


@mock.patch.object(Client, "request", return_value=({}, None))
def test_organizations_users_remove(mock_request):
    org_id = gen_uuid()
    user = "test@crate.io"

    call_command(
        "croud", "organizations", "users", "remove", "--user", user, "--org-id", org_id
    )
    assert_rest(
        mock_request,
        RequestMethod.DELETE,
        f"/api/v2/organizations/{org_id}/users/{user}/",
    )


def test_role_fqn_transform():
    user = {
        "organization_roles": [
            {"organization_id": "org-1", "role_fqn": "organization_admin"},
            {"organization_id": "org-2", "role_fqn": "organization_member"},
            {"organization_id": "org-3", "role_fqn": "organization_member"},
        ]
    }
    response = organization_role_fqn_transform(user["organization_roles"])
    assert response == "organization_admin"



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

import pytest

from croud.printer import (
    JsonFormatPrinter,
    TableFormatPrinter,
    WideTableFormatPrinter,
    YamlFormatPrinter,
    print_response,
)


@pytest.mark.parametrize(
    "rows,expected",
    (
        (None, "null"),
        ({}, "{}"),
        (
            {"a": "foo", "b": 1, "c": True, "d": {"x": 1, "y": "xyz", "z": False}},
            (
                "{\n"
                '  "a": "foo",\n'
                '  "b": 1,\n'
                '  "c": true,\n'
                '  "d": {\n'
                '    "x": 1,\n'
                '    "y": "xyz",\n'
                '    "z": false\n'
                "  }\n"
                "}"
            ),
        ),
    ),
)
def test_json_format(rows, expected):
    out = JsonFormatPrinter().format_rows(rows)
    assert out == expected


@pytest.mark.parametrize(
    "rows,expected",
    (
        (None, "null\n...\n"),
        ({}, "{}\n"),
        (
            {"a": "foo : bar", "b": 1, "c": True, "d": {"x": 1, "y": None, "z": False}},
            (
                "a: 'foo : bar'\n"
                "b: 1\n"
                "c: true\n"
                "d:\n"
                "  x: 1\n"
                "  y: null\n"
                "  z: false\n"
            ),
        ),
    ),
)
def test_yaml_format(rows, expected):
    out = YamlFormatPrinter().format_rows(rows)
    assert out == expected


@pytest.mark.parametrize(
    "rows,keys,transforms,expected",
    (
        (None, ["what", "ever"], None, ""),
        ({}, ["what", "ever"], None, ""),
        ([], [], None, ""),
        (
            [],
            ["what", "ever"],
            None,
            (
                "+--------+--------+\n"
                "| what   | ever   |\n"
                "|--------+--------|\n"
                "+--------+--------+"
            ),
        ),
        (
            {"a": "foo", "b": 1.234, "c": True, "d": "bar"},
            ["a", "b", "c"],
            {"a": str.upper, "b": int},
            (
                "+-----+-----+------+\n"
                "| a   |   b | c    |\n"
                "|-----+-----+------|\n"
                "| FOO |   1 | TRUE |\n"
                "+-----+-----+------+"
            ),
        ),
        (
            [
                {"a": "foo", "b": 1, "c": True},
                {"b": 2, "c": False, "a": {"bar": [{"value": 0}, {"value": 1}]}},
            ],
            None,
            None,
            (
                "+---------------------------------------+-----+-------+\n"
                "| a                                     |   b | c     |\n"
                "|---------------------------------------+-----+-------|\n"
                "| foo                                   |   1 | TRUE  |\n"
                '| {"bar": [{"value": 0}, {"value": 1}]} |   2 | FALSE |\n'
                "+---------------------------------------+-----+-------+"
            ),
        ),
        (
            [{"a": {"bar": "X"}, "b": [1, 2, 3]}, {"b": [2, 4, 8], "a": {"foo": "Y"}}],
            ["b", "a"],
            {"a": lambda field: field.get("bar", "N/A"), "b": sum},
            (
                "+-----+-----+\n"
                "|   b | a   |\n"
                "|-----+-----|\n"
                "|   6 | X   |\n"
                "|  14 | N/A |\n"
                "+-----+-----+"
            ),
        ),
        (
            {"k1": "v1", "k2": "v2", "k3": "v3"},
            ["k1", "k3"],
            None,
            (
                "+------+------+\n"
                "| k1   | k3   |\n"
                "|------+------|\n"
                "| v1   | v3   |\n"
                "+------+------+"
            ),
        ),
        (
            [
                {"k1": "v01", "k2": "v02", "k3": "v03"},
                {"k1": "v11", "k2": "v12", "k3": "v13"},
            ],
            ["k1", "k2"],
            None,
            (
                "+------+------+\n"
                "| k1   | k2   |\n"
                "|------+------|\n"
                "| v01  | v02  |\n"
                "| v11  | v12  |\n"
                "+------+------+"
            ),
        ),
        (
            [
                {"k1": "v01", "k2": "v02", "k3": "v03"},
                {"k1": "v11", "k2": "v12", "k3": "v13"},
            ],
            ["k1", "k2", "k4"],
            None,
            (
                "+------+------+\n"
                "| k1   | k2   |\n"
                "|------+------|\n"
                "| v01  | v02  |\n"
                "| v11  | v12  |\n"
                "+------+------+"
            ),
        ),
    ),
)
def test_tabular_format(rows, keys, transforms, expected):
    out = TableFormatPrinter(keys=keys, transforms=transforms).format_rows(rows)
    assert out == expected


@pytest.mark.parametrize(
    "rows,keys,transforms,expected",
    (
        (None, ["what", "ever"], None, ""),
        ({}, ["what", "ever"], None, ""),
        ([], [], None, ""),
        (
            [],
            ["what", "ever"],
            None,
            (
                "+--------+--------+\n"
                "| what   | ever   |\n"
                "|--------+--------|\n"
                "+--------+--------+"
            ),
        ),
        (
            {"a": "foo", "b": 1.234, "c": True, "d": "bar"},
            ["a", "b", "c"],
            {"a": str.upper, "b": int},
            (
                "+-----+-----+------+-----+\n"
                "| a   |   b | c    | d   |\n"
                "|-----+-----+------+-----|\n"
                "| FOO |   1 | TRUE | bar |\n"
                "+-----+-----+------+-----+"
            ),
        ),
    ),
)
def test_wide_format(rows, keys, transforms, expected):
    out = WideTableFormatPrinter(keys=keys, transforms=transforms).format_rows(rows)
    assert out == expected


def test_print_response(capsys):
    data = {"email": "test@crate.io", "username": "Google_1234"}
    errors = None
    output_fmt = "json"
    print_response(data, errors, output_fmt)
    output, _ = capsys.readouterr()
    assert "test@crate.io" in output
    assert "Google_1234" in output


def test_print_response_success_message(capsys):
    data = None
    errors = {}
    output_fmt = "json"
    success_message = "Deleted cluster."
    print_response(data, errors, output_fmt, success_message)
    _, err_output = capsys.readouterr()
    assert "Success" in err_output
    assert "Deleted cluster." in err_output


def test_print_response_error(capsys):
    data = None
    errors = {"message": "Resource not found.", "success": False}
    output_fmt = "table"
    print_response(data, errors, output_fmt)
    _, err_output = capsys.readouterr()
    assert "Error" in err_output
    assert "Resource not found." in err_output


def test_print_response_data_success(capsys):
    data = {"a": 1, "b": 2}
    errors = {}
    output_fmt = "json"
    success_message = "Posted numbers."
    print_response(data, errors, output_fmt, success_message)
    output, err_output = capsys.readouterr()
    assert '"a": 1' in output
    assert '"b": 2' in output
    assert "Success" in err_output
    assert "Posted numbers." in err_output



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

import pathlib
from unittest import mock

import pytest

from croud.config import WeakConfigProxy
from croud.config.configuration import Configuration
from croud.config.exceptions import InvalidConfiguration, InvalidProfile
from croud.config.util import clean_dict


@pytest.mark.parametrize(
    "source,cleaned",
    [
        ({"auth-token": "secret"}, {"auth-token": "xxxxxxxxxx"}),
        ({"auth-token": None}, {"auth-token": "xxxxxxxxxx"}),
        (
            {"profiles": [{"password": "secret"}]},
            {"profiles": [{"password": "xxxxxxxxxx"}]},
        ),
        (
            {"profiles": {"profile": {"secret": "secret"}}},
            {"profiles": {"profile": {"secret": "xxxxxxxxxx"}}},
        ),
    ],
)
def test_clean_dict(source, cleaned):
    assert clean_dict(source) == cleaned


def test_config_proxy_calls_ref_with_args_kwargs():
    getter = mock.Mock()

    proxy = WeakConfigProxy(getter, "arg", kw="kwarg")
    getter.assert_not_called()

    _ = proxy.data
    getter.assert_called_once_with("arg", kw="kwarg")


def test_config_proxy_gets_ref_attributes():
    m = mock.Mock()

    def getter():
        return m

    proxy = WeakConfigProxy(getter)
    assert proxy.data == m.data


def test_config_proxy_cannot_set_attribute():
    m = mock.Mock()

    def getter():
        return m

    proxy = WeakConfigProxy(getter)
    with pytest.raises(AttributeError, match="Cannot set attribute"):
        proxy.custom_attribute = "croud"


def test_default_configuration_instance(tmp_path):
    with mock.patch(
        "croud.config.configuration.user_config_dir", return_value=str(tmp_path)
    ):
        config = Configuration("test.yaml")
        assert config._file_path == tmp_path / "test.yaml"
        assert "aks1.westeurope.azure" in config.profiles
        assert "aks1.eastus2.azure" in config.profiles
        assert "eks1.eu-west-1.aws" in config.profiles
        assert config.name == "aks1.westeurope.azure"
        assert config.token is None
        assert config.endpoint == "https://console.cratedb.cloud"
        assert config.region == "aks1.westeurope.azure"
        assert config.format == "table"
        assert config.organization is None


def test_load_dump_configuration(tmpdir_factory):
    tmp_path = pathlib.Path(tmpdir_factory.getbasetemp()) / "foo" / "Crate"
    with mock.patch(
        "croud.config.configuration.user_config_dir", return_value=str(tmp_path)
    ):
        config = Configuration("test.yaml")
        # as long as we don't modify or dump the config, it's not written to disk
        assert config._file_path.exists() is False
        config.dump()
        assert str(config._config_dir) == str(tmp_path)
        assert tmp_path.stat().st_mode & 0o777 == 0o700  # only user access
        assert config._file_path.exists() is True
        assert config.config == config.load()


def test_load_invalid_configuration(tmp_path):
    with mock.patch(
        "croud.config.configuration.user_config_dir", return_value=str(tmp_path)
    ):
        with open(tmp_path / "test.yaml", "w") as fp:
            fp.write(
                """\
current-profile: foo
default-format: table
"""
            )
        config = Configuration("test.yaml")
        with pytest.raises(InvalidConfiguration, match="is not a valid configuration"):
            _ = config.config


def test_use_profile(config):
    config.use_profile("aks1.westeurope.azure")
    assert config.name == config.config["current-profile"] == "aks1.westeurope.azure"


def test_add_profile(config):
    config.add_profile("test", endpoint="http://localhost:8000", format="json")
    assert "test" in config.config["profiles"]
    assert config.config["profiles"]["test"] == {
        "auth-token": None,
        "endpoint": "http://localhost:8000",
        "format": "json",
    }


def test_add_profile_duplicate(config):
    with pytest.raises(InvalidProfile):
        config.add_profile("aks1.eastus2.azure", endpoint="http://localhost:8000")


def test_remove_profile(config):
    config.remove_profile("aks1.eastus2.azure")
    assert "aks1.eastus2.azure" not in config.config["profiles"]


def test_remove_profile_does_not_exist(config):
    with pytest.raises(InvalidProfile):
        config.remove_profile("foo")


def test_remove_profile_current(config):
    with pytest.raises(InvalidProfile):
        config.remove_profile(config.config["current-profile"])


def test_set_profile_options(config):
    profile = "test"
    config.add_profile(profile, endpoint="http://localhost:8000", format="table")

    config.set_auth_token(profile, "token")
    config.set_format(profile, "json")
    config.set_organization_id(profile, "53b70feb-1745-453a-b3f0-d0ac6ea5d55b")

    config.use_profile(profile)
    assert config.profile == {
        "endpoint": config.endpoint,
        "auth-token": config.token,
        "organization-id": config.organization,
        "format": config.format,
    }


def test_default_format(config):
    profile = "test"
    config.add_profile(profile, endpoint="http://localhost:8000")
    config.use_profile(profile)
    assert "format" not in config.profile
    assert config.format == config.config["default-format"]



import sys

from halo import Halo

# It's sadly not possible to set arbitrary colours, but cyan (the default)
# is close(ish) to Crate Blue (#00a6d1).
HALO = Halo(text="Working ...", spinner="dots", stream=sys.stderr)



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

from argparse import Namespace

from croud.api import Client
from croud.config import get_output_format
from croud.printer import print_response
from croud.util import require_confirmation


def clusters_get(args: Namespace) -> None:
    client = Client.from_args(args)
    data, errors = client.get(f"/api/v2/clusters/{args.id}/")
    print_response(
        data=data, errors=errors, output_fmt=get_output_format(args),
    )


def clusters_list(args: Namespace) -> None:
    params = {}
    if args.project_id:
        params["project_id"] = args.project_id

    client = Client.from_args(args)
    data, errors = client.get("/api/v2/clusters/", params=params)
    print_response(
        data=data,
        errors=errors,
        keys=[
            "id",
            "name",
            "num_nodes",
            "crate_version",
            "project_id",
            "username",
            "fqdn",
            "channel",
        ],
        output_fmt=get_output_format(args),
    )


def clusters_deploy(args: Namespace) -> None:
    body = {
        "crate_version": args.version,
        "name": args.cluster_name,
        "password": args.password,
        "product_name": args.product_name,
        "product_tier": args.tier,
        "project_id": args.project_id,
        "username": args.username,
        "channel": args.channel,
    }
    if args.unit:
        body["product_unit"] = args.unit
    _handle_edge_params(body, args)
    client = Client.from_args(args)
    data, errors = client.post("/api/v2/clusters/", body=body)
    print_response(
        data=data,
        errors=errors,
        keys=["id", "name", "fqdn", "url"],
        success_message=(
            "Cluster deployed. It may take a few minutes to complete the changes."
        ),
        output_fmt=get_output_format(args),
    )


def clusters_scale(args: Namespace) -> None:
    body = {"product_unit": args.unit}
    client = Client.from_args(args)
    data, errors = client.put(f"/api/v2/clusters/{args.cluster_id}/scale/", body=body)
    print_response(
        data=data,
        errors=errors,
        keys=["id", "name", "num_nodes"],
        success_message=(
            "Cluster scaled. It may take a few minutes to complete the changes."
        ),
        output_fmt=get_output_format(args),
    )


def clusters_upgrade(args: Namespace) -> None:
    body = {"crate_version": args.version}
    client = Client.from_args(args)
    data, errors = client.put(f"/api/v2/clusters/{args.cluster_id}/upgrade/", body=body)
    print_response(
        data=data,
        errors=errors,
        keys=["id", "name", "crate_version"],
        success_message=(
            "Cluster upgraded. It may take a few minutes to complete the changes."
        ),
        output_fmt=get_output_format(args),
    )


@require_confirmation(
    "Are you sure you want to delete the cluster?",
    cancel_msg="Cluster deletion cancelled.",
)
def clusters_delete(args: Namespace) -> None:
    client = Client.from_args(args)
    data, errors = client.delete(f"/api/v2/clusters/{args.cluster_id}/")
    print_response(
        data=data,
        errors=errors,
        success_message="Cluster deleted.",
        output_fmt=get_output_format(args),
    )


def clusters_restart_node(args: Namespace) -> None:
    client = Client.from_args(args)
    data, errors = client.delete(
        f"/api/v2/clusters/{args.cluster_id}/nodes/{args.ordinal}"
    )
    print_response(
        data=data,
        errors=errors,
        keys=["code", "status"],
        success_message=(
            "Node restarted. It may take a few minutes to complete the changes."
        ),
        output_fmt=get_output_format(args),
    )


def clusters_set_deletion_protection(args: Namespace) -> None:
    body = {"deletion_protected": args.value}
    client = Client.from_args(args)
    data, errors = client.put(
        f"/api/v2/clusters/{args.cluster_id}/deletion-protection/", body=body
    )
    print_response(
        data=data,
        errors=errors,
        keys=["id", "name", "deletion_protected"],
        success_message=("Cluster deletion protection status successfully updated"),
        output_fmt=get_output_format(args),
    )


# We want to map the custom hardware specs to slightly nicer params in croud,
# hence this mapping here
def _handle_edge_params(body, args):
    if args.cpus:
        body.setdefault("hardware_specs", {})["cpus_per_node"] = args.cpus
    if args.disks:
        body.setdefault("hardware_specs", {})["disks_per_node"] = args.disks
    if args.disk_size_gb:
        body.setdefault("hardware_specs", {})["disk_size_per_node_bytes"] = (
            args.disk_size_gb * 1024 * 1024 * 1024
        )
    if args.disk_type:
        body.setdefault("hardware_specs", {})["disk_type"] = args.disk_type
    if args.memory_size_mb:
        body.setdefault("hardware_specs", {})["memory_per_node_bytes"] = (
            args.memory_size_mb * 1024 * 1024
        )



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

import pytest
import yaml
from marshmallow.exceptions import ValidationError

from croud.config.configuration import ConfigSchema, ProfileSchema


@pytest.mark.parametrize(
    "data",
    [
        """\
    auth-token: NULL
    endpoint: http://localhost:8000
    """,
        """\
    auth-token: NULL
    endpoint: http://localhost:8000
    format: json
    """,
        """\
    auth-token: NULL
    endpoint: http://localhost:8000
    organization-id: 804f1f14-0a5e-46b4-ab37-6d046b1fa07b
    """,
    ],
)
def test_load_profile_schema_valid(data):
    schema = ProfileSchema()
    schema.load(yaml.safe_load(data))


@pytest.mark.parametrize(
    "data,errors",
    [
        (
            """\
    auth-token: NULL
    endpoint: http://localhost:8000
    invalid: true
    """,
            {"invalid": ["Unknown field."]},
        ),
        (
            """\
    format: json
    organization-id: 804f1f14-0a5e-46b4-ab37-6d046b1fa07b
    """,
            {
                "auth-token": ["Missing data for required field."],
                "endpoint": ["Missing data for required field."],
            },
        ),
        (
            """\
    auth-token: NULL
    endpoint: http://localhost:8000
    format: invalid
    """,
            {"format": ["Must be one of: table, wide, json, yaml."]},
        ),
    ],
)
def test_load_profile_schema_invalid(data, errors):
    schema = ProfileSchema()
    with pytest.raises(ValidationError) as exc_info:
        schema.load(yaml.safe_load(data))
    assert exc_info.value.messages == errors


@pytest.mark.parametrize(
    "data",
    [
        """\
    current-profile: local
    default-format: table
    profiles:
      local:
        endpoint: https://localhost:8000
        auth-token: NULL
    """,
    ],
)
def test_load_config_schema_valid(data):
    schema = ConfigSchema()
    schema.load(yaml.safe_load(data))


@pytest.mark.parametrize(
    "data,errors",
    [
        (
            """\
    current-profile: local
    default-format: table
    profiles: {}
    """,
            {"_schema": ["The 'current-profile' does not exist in 'profiles'."]},
        ),
        (
            """\
    {}
    """,
            {
                "current-profile": ["Missing data for required field."],
                "default-format": ["Missing data for required field."],
                "profiles": ["Missing data for required field."],
            },
        ),
    ],
)
def test_load_config_schema_invalid(data, errors):
    schema = ConfigSchema()
    with pytest.raises(ValidationError) as exc_info:
        schema.load(yaml.safe_load(data))
    assert exc_info.value.messages == errors



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

from argparse import Namespace

from croud.api import Client
from croud.config import get_output_format
from croud.printer import print_error, print_response
from croud.util import org_id_config_fallback, require_confirmation


@org_id_config_fallback
def project_create(args: Namespace) -> None:
    body = {"name": args.name, "organization_id": args.org_id}
    _handle_custom_backups(body, args)
    client = Client.from_args(args)
    data, errors = client.post("/api/v2/projects/", body=body)
    print_response(
        data=data,
        errors=errors,
        keys=["id"],
        success_message="Project created.",
        output_fmt=get_output_format(args),
    )


@require_confirmation(
    "Are you sure you want to delete the project?",
    cancel_msg="Project deletion cancelled.",
)
def project_delete(args: Namespace) -> None:
    client = Client.from_args(args)
    data, errors = client.delete(f"/api/v2/projects/{args.project_id}/")
    print_response(
        data=data,
        errors=errors,
        success_message="Project deleted.",
        output_fmt=get_output_format(args),
    )


def project_edit(args: Namespace) -> None:
    client = Client.from_args(args)
    body = {}
    if args.name:
        body["name"] = args.name
    if not body:
        print_error("No input arguments found.")
        exit(1)

    data, errors = client.patch(f"/api/v2/projects/{args.project_id}/", body=body)
    print_response(
        data=data,
        errors=errors,
        keys=["id", "name"],
        success_message="Project edited.",
        output_fmt=get_output_format(args),
    )


def projects_get(args: Namespace) -> None:
    client = Client.from_args(args)
    data, errors = client.get(f"/api/v2/projects/{args.id}/")
    print_response(
        data=data,
        errors=errors,
        output_fmt=get_output_format(args),
        transforms={"backup_location": _transform_backup_location},
    )


def projects_list(args: Namespace) -> None:
    client = Client.from_args(args)
    data, errors = client.get("/api/v2/projects/")
    print_response(
        data=data,
        errors=errors,
        keys=["id", "name", "region", "organization_id", "backup_location"],
        transforms={"backup_location": _transform_backup_location},
        output_fmt=get_output_format(args),
    )


def _handle_custom_backups(body, args: Namespace) -> None:
    if args.backup_location_type:
        body.setdefault("backup_location", {})[
            "location_type"
        ] = args.backup_location_type
    if args.backup_location:
        body.setdefault("backup_location", {})["location"] = args.backup_location
    if args.backup_location_access_key_id and args.backup_location_secret_access_key:
        body.setdefault("backup_location", {})["credentials"] = {
            "access_key_id": args.backup_location_access_key_id,
            "secret_access_key": args.backup_location_secret_access_key,
        }
    if args.backup_location_endpoint_url:
        body.setdefault("backup_location", {})["additional_config"] = {
            "endpoint_url": args.backup_location_endpoint_url,
        }


def _transform_backup_location(field):
    if not field:
        return "default"

    return f"{field['location_type']}://{field['location']}"



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

import difflib
import re
import sys
import uuid
from doctest import _ellipsis_match  # type: ignore
from functools import partial
from unittest import mock

from croud.__main__ import main
from croud.api import RequestMethod

normalize = partial(re.sub, r"\s+", "")

UNDEFINED = object()


def ndiff(actual: str, expected: str) -> str:
    engine = difflib.Differ(charjunk=difflib.IS_CHARACTER_JUNK)
    diff = list(engine.compare(expected.splitlines(True), actual.splitlines(True)))
    return "\n".join(line.rstrip() for line in diff)


def assert_ellipsis_match(actual: str, expected: str) -> None:
    diff = ndiff(expected.strip(), actual.strip())
    equality = _ellipsis_match(normalize(expected.strip()), normalize(actual.strip()))
    assert equality, diff


def call_command(*argv):
    with mock.patch.object(sys, "argv", argv):
        main()


def assert_rest(
    mock_request, method: RequestMethod, endpoint, *, params=None, body=None
):
    args = [method, endpoint]
    kwargs = {}
    if method != RequestMethod.GET and body is not UNDEFINED:
        kwargs["body"] = body
    if params is not UNDEFINED:
        kwargs["params"] = params

    mock_request.assert_called_once_with(*args, **kwargs)


def gen_uuid():
    return str(uuid.uuid4())



#!/usr/bin/env python
#
# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

from pkg_resources.extern.packaging.version import Version
from setuptools import find_packages, setup

__version__ = Version("0.29.0")

try:
    with open("README.rst", "r", encoding="utf-8") as f:
        readme = f.read()
except IOError:
    readme = ""


setup(
    name="croud",
    author="Crate.io",
    author_email="office@crate.io",
    url="https://github.com/crate/croud",
    description="A command line interface for CrateDB Cloud",
    long_description=readme,
    version=str(__version__),
    entry_points={"console_scripts": ["croud = croud.__main__:main"]},
    packages=find_packages(),
    install_requires=[
        "appdirs==1.4.3",
        "certifi",
        "colorama==0.4.3",
        "marshmallow==3.5.1",
        "pyyaml==5.4",
        "requests==2.23.0",
        "tabulate==0.8.6",
        "yarl==1.4.2",
        "halo==0.0.31",
    ],
    extras_require={
        "testing": ["tox==3.14.2"],
        "development": [
            "black==19.10b0",
            "flake8==3.7.9",
            "isort==4.3.21",
            "mypy==0.812",
        ],
    },
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Topic :: Database",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
)



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

from marshmallow import Schema, fields, post_load
from marshmallow.exceptions import ValidationError
from marshmallow.validate import OneOf

OUTPUT_FORMATS = ("table", "wide", "json", "yaml")  # we want to keep them sorted!


class ProfileSchema(Schema):
    """
    Schema for profile entries
    """

    auth_token = fields.String(
        attribute="auth-token", data_key="auth-token", required=True, allow_none=True
    )
    endpoint = fields.String(required=True)
    format = fields.String(validate=OneOf(choices=OUTPUT_FORMATS))
    organization_id = fields.String(
        attribute="organization-id", data_key="organization-id", allow_none=True
    )
    region = fields.String(required=False)


class ConfigSchema(Schema):
    """
    Schema for configuration file
    """

    default_format = fields.String(
        attribute="default-format", data_key="default-format", required=True
    )
    current_profile = fields.String(
        attribute="current-profile", data_key="current-profile", required=True
    )
    profiles = fields.Dict(
        keys=fields.String(), values=fields.Nested(ProfileSchema()), required=True,
    )

    @post_load
    def validate_current_profile(self, data, **kwargs):
        if data["current-profile"] not in data["profiles"]:
            raise ValidationError("The 'current-profile' does not exist in 'profiles'.")
        return data



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

from argparse import Namespace

from croud.api import Client
from croud.config import get_output_format
from croud.printer import print_response


def project_users_add(args: Namespace) -> None:
    client = Client.from_args(args)

    data, errors = client.post(
        f"/api/v2/projects/{args.project_id}/users/",
        body={"user": args.user, "role_fqn": args.role},
    )
    if data is not None and data.get("added", False):
        success_message = "User added to project."
    else:
        success_message = "Role altered for user."

    print_response(
        data=data,
        errors=errors,
        keys=["user_id", "project_id", "role_fqn"],
        success_message=success_message,
        output_fmt=get_output_format(args),
    )


def role_fqn_transform(field):
    return field[0]["role_fqn"]


def project_users_list(args: Namespace) -> None:
    client = Client.from_args(args)
    data, errors = client.get(f"/api/v2/projects/{args.project_id}/users/")
    print_response(
        data=data,
        errors=errors,
        output_fmt=get_output_format(args),
        keys=["uid", "email", "username", "project_roles"],
        transforms={"project_roles": role_fqn_transform},
    )


def project_users_remove(args: Namespace) -> None:
    client = Client.from_args(args)
    data, errors = client.delete(
        f"/api/v2/projects/{args.project_id}/users/{args.user}/"
    )
    print_response(
        data=data,
        errors=errors,
        success_message="User removed from project.",
        output_fmt=get_output_format(args),
    )



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

from unittest import mock

import pytest

from croud.api import Client, RequestMethod
from croud.printer import print_raw
from tests.util import assert_rest, call_command

EDGE_REGION_LIST = [
    {
        "dc": None,
        "deprecated": False,
        "description": "region-description",
        "is_edge_region": True,
        "last_seen": None,
        "name": "region-name",
        "organization_id": "organization-id",
        "status": "DOWN",
    }
]


@mock.patch.object(Client, "request", return_value=({}, None))
def test_regions_list(mock_request):
    call_command("croud", "regions", "list")
    assert_rest(mock_request, RequestMethod.GET, "/api/v2/regions/")


@mock.patch.object(Client, "request", return_value=({"name": "region-name"}, None))
def test_regions_create_all_params(mock_request):
    call_command(
        "croud",
        "regions",
        "create",
        "--description",
        "region-description",
        "--aws-bucket",
        "backup-bucket",
        "--aws-region",
        "backup-region",
        "--provider",
        "EDGE",
        "--org-id",
        "organization-id",
        "--name",
        "region-name",
        "--yes",
    )

    mock_request.call_args_list[0].assert_called_with(
        RequestMethod.POST,
        "/api/v2/regions/",
        body={
            "description": "region-description",
            "aws_bucket": "backup-bucket",
            "aws_region": "backup-region",
            "provider": "EDGE",
            "name": "region-name",
            "organization_id": "organization-id",
        },
        params=None,
    )
    mock_request.call_args_list[1].assert_called_with(
        RequestMethod.GET, "/api/v2/regions/region-name/install-token/", params=None
    )


@mock.patch.object(Client, "request", return_value=({"name": "region-name"}, None))
def test_regions_create_mandatory_params(mock_request):
    call_command(
        "croud",
        "regions",
        "create",
        "--description",
        "region-description",
        "--provider",
        "EDGE",
        "--yes",
    )

    mock_request.call_args_list[0].assert_called_with(
        RequestMethod.POST,
        "/api/v2/regions/",
        body={"description": "region-description", "provider": "EDGE"},
        params=None,
    )
    mock_request.call_args_list[1].assert_called_with(
        RequestMethod.GET, "/api/v2/regions/region-name/install-token/", params=None
    )


@mock.patch("croud.regions.commands.print_raw", wraps=print_raw)
@mock.patch.object(
    Client,
    "request",
    side_effect=[({"name": "region-name"}, None), ({"token": "my-token"}, None)],
)
def test_regions_create_install_command(mock_request, mock_raw_printer, client):
    call_command(
        "croud",
        "regions",
        "create",
        "--description",
        "region-description",
        "--provider",
        "EDGE",
        "--yes",
    )

    mock_request.call_args_list[0].assert_called_with(
        RequestMethod.POST,
        "/api/v2/regions/",
        body={"description": "region-description", "provider": "EDGE"},
        params=None,
    )
    mock_request.call_args_list[1].assert_called_with(
        RequestMethod.GET, "/api/v2/regions/region-name/install-token/", params=None
    )

    assert any(
        f"$ bash <( wget -qO- {client.base_url}/edge/cratedb-cloud-edge.sh) my-token"
        in item
        for item in mock_raw_printer.call_args[0][0]
    )


@mock.patch.object(Client, "request", return_value=({}, None))
def test_regions_create_missing_description(mock_request, capsys):
    with pytest.raises(SystemExit):
        call_command(
            "croud", "regions", "create", "--provider", "EDGE", "--yes",
        )
    mock_request.assert_not_called()
    _, err_output = capsys.readouterr()
    assert "The following arguments are required: --description" in err_output


@mock.patch.object(Client, "request", return_value=(None, {}))
def test_regions_create_aborted(mock_request, capsys):
    with mock.patch("builtins.input", side_effect=["Nooooo"]) as mock_input:
        call_command(
            "croud",
            "regions",
            "create",
            "--description",
            "region-description",
            "--provider",
            "EDGE",
        )
    mock_request.assert_not_called()
    mock_input.assert_called_once_with(
        "Creating a region is an experimental feature. Are you sure you want to proceed? [yN] "  # noqa
    )

    _, err_output = capsys.readouterr()
    assert "Region creation cancelled." in err_output


@mock.patch.object(
    Client, "request", side_effect=[(EDGE_REGION_LIST, None), ({}, None)],
)
def test_regions_delete(mock_request, capsys):
    call_command(
        "croud", "regions", "delete", "-y", "--name", "region-name",
    )
    mock_request.assert_has_calls(
        [
            mock.call(RequestMethod.GET, "/api/v2/regions/", params=None),
            mock.call(
                RequestMethod.DELETE,
                "/api/v2/regions/region-name/",
                params=None,
                body=None,
            ),
        ]
    )


@mock.patch.object(
    Client,
    "request",
    side_effect=[
        (EDGE_REGION_LIST, None),
        (
            None,
            {
                "success": False,
                "message": "Deletion requires no resources to be related to the target.",  # noqa
                "related_resources": {
                    "projects": [
                        "5d8e1a81-e2fe-48bb-8cc2-fbdbebfe2d8f",
                        "42023d1f-462e-4130-aa82-6025342b11c5",
                    ]
                },
            },
        ),
    ],
)
def test_regions_delete_with_projects_fails(mock_request, capsys):

    call_command(
        "croud", "regions", "delete", "-y", "--name", "region-name",
    )
    mock_request.assert_has_calls(
        [
            mock.call(RequestMethod.GET, "/api/v2/regions/", params=None),
            mock.call(
                RequestMethod.DELETE,
                "/api/v2/regions/region-name/",
                params=None,
                body=None,
            ),
        ]
    )
    data, err_output = capsys.readouterr()
    assert "Deletion requires no resources to be related to the target." in err_output
    assert "5d8e1a81-e2fe-48bb-8cc2-fbdbebfe2d8f" in data
    assert "42023d1f-462e-4130-aa82-6025342b11c5" in data


@mock.patch.object(
    Client, "request", side_effect=[({}, None)],
)
def test_regions_delete_missing_name(mock_request, capsys):
    with pytest.raises(SystemExit):
        call_command(
            "croud", "regions", "delete", "-y",
        )
    mock_request.assert_not_called()
    _, err_output = capsys.readouterr()
    assert "The following arguments are required: --name" in err_output


@mock.patch.object(
    Client,
    "request",
    side_effect=[
        (
            [
                {
                    "dc": None,
                    "deprecated": False,
                    "description": "region-description",
                    "is_edge_region": True,
                    "last_seen": None,
                    "name": "region-name",
                    "organization_id": "organization-id",
                    "status": "UP",
                }
            ],
            None,
        ),
    ],
)
def test_regions_delete_status_up(mock_request, capsys, client):

    call_command("croud", "regions", "delete", "-y", "--name", "region-name")
    mock_request.assert_called_once_with(
        RequestMethod.GET, "/api/v2/regions/", params=None,
    )
    data, err_output = capsys.readouterr()
    msg = (
        "Your region is still connected to CrateDB Cloud. Please uninstall "
        "the CrateDB Edge stack from the region before deleting it by "
        "running the script below:\nbash <(wget -qO- "
        f"{client.base_url}/edge/uninstall-cratedb-cloud-edge.sh)"
    )
    assert msg in err_output


@mock.patch.object(
    Client, "request", side_effect=[(EDGE_REGION_LIST, None), ({}, None)],
)
def test_regions_delete_status_down(mock_request, capsys):

    call_command("croud", "regions", "delete", "-y", "--name", "region-name")
    mock_request.assert_has_calls(
        [
            mock.call(RequestMethod.GET, "/api/v2/regions/", params=None),
            mock.call(
                RequestMethod.DELETE,
                "/api/v2/regions/region-name/",
                params=None,
                body=None,
            ),
        ]
    )


@mock.patch.object(
    Client, "request", side_effect=[({}, None)],
)
def test_regions_delete_bad_name(mock_request, capsys):

    region_name = "region-name"

    call_command("croud", "regions", "delete", "-y", "--name", region_name)
    mock_request.assert_called_once_with(
        RequestMethod.GET, "/api/v2/regions/", params=None,
    )
    _, err_output = capsys.readouterr()
    assert f"The region {region_name} does not exist." in err_output
